"""Tests for the bumpversion.click_config module."""

from pathlib import Path
from typing import Optional

import click
import httpx
import pytest

from bumpversion.click_config import config_option, download_url, resolve_conf_location
from bumpversion.exceptions import BadInputError


@click.command()
@config_option()
def hello(config: Optional[Path]):
    """Say hello."""
    click.echo(f"Hello {config}!")
    if config:
        click.echo(f"path exists: {config.exists()}")
    else:
        click.echo("Nothing was passed.")


class TestConfigOption:
    """Tests to for config_option."""

    def test_passing_missing_file_raises_error(self, runner):
        """Explicitly passing a path that does not exist should raise an error."""
        result = runner.invoke(hello, ["--config", "idont/exist.txt"])

        assert result.exit_code != 0

    def test_passing_nothing_returns_none(self, runner):
        """If nothing is passed, the value is `None`."""
        result = runner.invoke(hello, [])

        assert result.exit_code == 0
        assert "Hello None!" in result.output

    def test_passing_existing_file_returns_path(self, runner, fixtures_path: Path):
        """Passing an existing file path should return a Path object."""
        config_path = fixtures_path / "basic_cfg.toml"
        result = runner.invoke(hello, ["--config", str(config_path)])

        assert result.exit_code == 0
        assert f"Hello {config_path}!" in result.output

    def test_passing_url_returns_path(self, runner, fixtures_path: Path, httpserver):
        """Passing an existing file path should return a Path object."""
        # Assemble
        config = fixtures_path.joinpath("basic_cfg.toml").read_text()
        httpserver.serve_content(config)

        # Act
        result = runner.invoke(hello, ["--config", f"{httpserver.url}/basic_cfg.toml"])

        # Assert
        assert result.exit_code == 0
        assert "path exists: True" in result.output


class TestResolveConfLocation:
    """Tests for resolve_conf_location."""

    def test_resolves_file_path(self, fixtures_path: Path):
        """Should return a Path object for a valid local file path."""
        dummy_path = fixtures_path / "basic_cfg.toml"
        result = resolve_conf_location(str(dummy_path))
        assert isinstance(result, Path)
        assert result == dummy_path

    def test_resolves_valid_url(self, fixtures_path: Path, httpserver):
        """Should download the file if given a valid URL."""
        # Arrange
        config = fixtures_path.joinpath("basic_cfg.toml").read_text()
        httpserver.serve_content(config)
        url = f"{httpserver.url}/basic_cfg.toml"

        # Act
        result = resolve_conf_location(url)

        # Assert
        assert isinstance(result, Path)
        assert result.read_text() == config


class TestDownloadUrl:
    """Tests for download_url."""

    def test_download_url_success(self, fixtures_path: Path, httpserver):
        """Should return a Path for a valid request."""
        # Arrange
        config = fixtures_path.joinpath("basic_cfg.toml").read_text()
        httpserver.serve_content(config)
        url = f"{httpserver.url}/basic_cfg.toml"

        # Act
        result = download_url(url)

        # Assert
        assert isinstance(result, Path)
        assert result.read_text() == config

    def test_download_url_request_error(self, mocker):
        """Should raise BadInputError for a RequestError."""
        # Arrange
        mocker.patch("httpx.get", side_effect=httpx.RequestError("Request failed"))

        # Act & Assert
        with pytest.raises(BadInputError, match="Unable to download configuration from URL"):
            download_url("http://example.com/config.txt")

    def test_download_url_http_status_error(self, mocker):
        """Should raise BadInputError for an HTTPStatusError."""
        mock_response = mocker.Mock(spec=httpx.Response)
        mock_response.status_code = 404
        mock_response.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Error", request=mocker.Mock(), response=mock_response
        )
        mocker.patch("httpx.get", return_value=mock_response)

        # Act & Assert
        with pytest.raises(BadInputError, match="Error response 404 while requesting"):
            download_url("http://example.com/config.txt")



"""Tests for the mercurial SCM."""

import json
import shutil
import subprocess
from pathlib import Path
from unittest.mock import patch

import pytest
from pytest import param

from bumpversion import cli
from bumpversion.scm.hg import Mercurial, tag, assert_nondirty
from bumpversion.scm.models import LatestTagInfo, SCMConfig
from bumpversion.exceptions import BumpVersionError, DirtyWorkingDirectoryError
from bumpversion.utils import run_command
from tests.conftest import inside_dir

BASIC_CONFIG = """
[tool.bumpversion]
commit = true
tag = true
current_version = "1.0.0"
parse = "(?P<major>\\\\d+)\\\\.(?P<minor>\\\\d+)\\\\.(?P<patch>\\\\d+)(\\\\-(?P<release>[a-z]+))?"
serialize = [
    "{major}.{minor}.{patch}-{release}",
    "{major}.{minor}.{patch}"
]
"""


@pytest.fixture
def hg_instance(scm_config: SCMConfig):
    """Return a Mercurial instance."""
    return Mercurial(scm_config)


def commit_readme(message: str) -> str:
    """Commit the readme file and return the commit sha1, so we can tag it later on."""
    subprocess.run(["hg", "add", "readme.md"], check=True)  # noqa: S603, S607
    subprocess.run(["hg", "commit", "-m", message], check=True)  # noqa: S603, S607
    result = subprocess.run(  # noqa: S603
        ["hg", "id", "-r", "tip"], capture_output=True, check=True, text=True  # noqa: S607
    )
    return result.stdout.strip()


class TestMercurial:
    """Tests of the Mercurial class."""

    class TestIsAvailable:
        """Tests of the is_available method."""

        def test_hg_is_not_available(self, tmp_path: Path, hg_instance: Mercurial) -> None:
            """Should return false if it is not a mercurial repo."""
            with inside_dir(tmp_path):
                assert not hg_instance.is_available()

        def test_hg_is_available(self, hg_repo: Path, hg_instance: Mercurial) -> None:
            """Should return false if it is not a mercurial repo."""
            with inside_dir(hg_repo):
                assert hg_instance.is_available()

    class TestLatestTagInfo:
        """Test for the Mercurial.latest_tag_info() function."""

        def test_returns_default_if_not_available(self, tmp_path: Path, hg_instance: Mercurial) -> None:
            """If git is not available, it returns the default LatestTagInfo object, with all None values."""
            # Arrange
            expected = LatestTagInfo()

            # Act
            with inside_dir(tmp_path):
                info = hg_instance.latest_tag_info()

            # Assert
            assert info == expected

        def test_returns_correct_commit_and_tag_info(self, hg_repo: Path, hg_instance: Mercurial) -> None:
            """Should return information that it is a git repo."""
            # Arrange
            tag_prefix = "app/"
            readme = hg_repo.joinpath("readme.md")
            readme.touch()
            hg_instance.config.tag_name = f"{tag_prefix}{{new_version}}"

            # Act
            with inside_dir(hg_repo):
                commit_readme("first")
                tag(f"{tag_prefix}0.1.0", sign=False, message="bumpversion")
                tag_info = hg_instance.latest_tag_info()

            # Assert
            assert tag_info.commit_sha is not None
            assert tag_info.current_version == "0.1.0"
            assert tag_info.current_tag == f"{tag_prefix}0.1.0"
            assert tag_info.distance_to_latest_tag == 0
            assert tag_info.branch_name == "default"
            assert tag_info.short_branch_name == "default"
            assert tag_info.repository_root == hg_repo
            assert tag_info.dirty is False

    class TestAddPath:
        """Tests for the add_path method in the Mercurial class."""

        def test_adds_to_repo_if_file_is_within_repo(self, hg_repo: Path, hg_instance: Mercurial):
            """The file is added to the repo when it is within the repo root."""
            # Arrange
            new_file = hg_repo / "newfile.txt"
            new_file.write_text("new file")

            # Act
            with inside_dir(hg_repo):
                hg_instance.add_path(new_file)
                result = run_command(["hg", "status"])

            # Assert
            assert result.stdout == "A newfile.txt\n"

        @patch("bumpversion.scm.hg.run_command")
        def test_does_not_add_path_when_not_in_repository(
            self, mock_run_command, hg_repo: Path, hg_instance: Mercurial, fixtures_path: Path
        ):
            """The file is not added to the repo when it is outside the repo root."""
            # Arrange
            new_file = fixtures_path / "basic_cfg.toml"

            # Act
            with inside_dir(hg_repo):
                hg_instance.add_path(new_file)

            mock_run_command.assert_called_once_with(["hg", "root"])

    class TestGetAllTags:
        """Tests for the get_all_tags method."""

        def test_returns_correct_tags(self, hg_repo: Path, hg_instance: Mercurial):
            """Returns all the tags in the repo."""
            # Arrange
            readme = hg_repo.joinpath("readme.md")
            readme.touch()
            file1 = hg_repo.joinpath("file1.txt")
            file1.touch()
            file2 = hg_repo.joinpath("file2.txt")
            file2.touch()
            with inside_dir(hg_repo):
                commit_readme("first")
                tag("tag1")
                hg_instance.add_path(file1)
                run_command(["hg", "commit", "-m", "add file1"])
                tag("tag2")
                hg_instance.add_path(file2)
                run_command(["hg", "commit", "-m", "add file2"])
                tag("tag3")

                # Act
                tags = hg_instance.get_all_tags()

            # Assert
            assert set(tags) == {"tip", "tag1", "tag2", "tag3"}

        def test_handles_empty_tags(self, hg_repo: Path, hg_instance: Mercurial):
            """When there are no tags, an empty list is returned."""
            # Arrange
            with inside_dir(hg_repo):
                hg_repo.joinpath("readme.md").touch()
                commit_readme("first")

                # Act
                tags = hg_instance.get_all_tags()

            # Assert
            assert set(tags) == {"tip"}

        @pytest.mark.parametrize(
            ["side_effect"],
            [
                param(subprocess.CalledProcessError(1, "cmd"), id="command failure"),
                param(FileNotFoundError, id="file not found"),
                param(PermissionError, id="permission error"),
            ],
        )
        def test_failure_raises_error(self, hg_instance, side_effect, mocker):
            """On failure, it raises a BumpVersionError exception."""
            # Arrange
            mocker.patch("bumpversion.utils.run_command", side_effect=side_effect)

            # Act, Assert
            with pytest.raises(BumpVersionError):
                hg_instance.get_all_tags()


@pytest.mark.skipif(not shutil.which("hg"), reason="Mercurial is not available.")
def test_hg_bump_version_commits_changes(hg_repo: Path, fixtures_path: Path, runner) -> None:
    """Should commit changes to the repo when configured."""
    dest_cfg = hg_repo / ".bumpversion.toml"
    dest_cfg.write_text(BASIC_CONFIG)

    with inside_dir(hg_repo):
        run_command(["hg", "add", ".bumpversion.toml"])
        run_command(["hg", "commit", "-m", "add configuration"])
        runner.invoke(
            cli.cli,
            ["bump", "minor", "-vv"],
        )
        hg_result = run_command(["hg", "tags", "-T", "json"])

    hg_logs = json.loads(hg_result.stdout)
    tags = [item["tag"] for item in hg_logs]
    assert "v1.1.0" in tags


class TestAssertNonDirty:
    """Tests for the assert_nondirty function."""

    def test_recognizes_clean_working_directory(self, hg_repo: Path):
        """A clean repo does not raise an exception."""
        with inside_dir(hg_repo):
            assert_nondirty()

    def test_dirty_working_directory_raises_exception(self, hg_repo: Path):
        """A dirty working directory raises an exception."""
        # Arrange
        readme = hg_repo.joinpath("readme.md")
        readme.touch()
        with inside_dir(hg_repo):
            commit_readme("first")
            readme.write_text("new text")
            with pytest.raises(DirtyWorkingDirectoryError) as exc:
                assert_nondirty()

        assert "Mercurial working directory is not clean" in str(exc.value)
        assert "M readme.md" in str(exc.value)



"""General utilities."""

from __future__ import annotations

import string
import subprocess
from pathlib import Path
from subprocess import CompletedProcess
from typing import Any, List, Optional, Tuple, Union

from bumpversion.exceptions import BumpVersionError
from bumpversion.ui import get_indented_logger

logger = get_indented_logger(__name__)


Pathlike = Union[str, Path]
"""A type alias for a path-like object."""


def extract_regex_flags(regex_pattern: str) -> Tuple[str, str]:
    """
    Extract the regex flags from the regex pattern.

    Args:
        regex_pattern: The pattern that might start with regex flags

    Returns:
        A tuple of the regex pattern without the flag string and regex flag string
    """
    import re

    flag_pattern = r"^(\(\?[aiLmsux]+\))"
    bits = re.split(flag_pattern, regex_pattern)
    return (regex_pattern, "") if len(bits) == 1 else (bits[2], bits[1])


def recursive_sort_dict(input_value: Any) -> Any:
    """Sort a dictionary recursively."""
    if not isinstance(input_value, dict):
        return input_value

    return {key: recursive_sort_dict(input_value[key]) for key in sorted(input_value.keys())}


def key_val_string(d: dict) -> str:
    """Render the dictionary as a comma-delimited key=value string."""
    return ", ".join(f"{k}={v}" for k, v in sorted(d.items()))


def labels_for_format(serialize_format: str) -> List[str]:
    """Return a list of labels for the given serialize_format."""
    return [item[1] for item in string.Formatter().parse(serialize_format) if item[1]]


def get_overrides(**kwargs) -> dict:
    """Return a dictionary containing only the overridden key-values."""
    return {key: val for key, val in kwargs.items() if val is not None}


def get_nested_value(d: dict, path: str) -> Any:
    """
    Retrieves the value of a nested key in a dictionary based on the given path.

    Args:
        d: The dictionary to search.
        path: A string representing the path to the nested key, separated by periods.

    Returns:
        The value of the nested key.

    Raises:
        KeyError: If a key in the path does not exist.
        ValueError: If an element in the path is not a dictionary.
    """
    keys = path.split(".")
    current_element = d

    for key in keys:
        if not isinstance(current_element, dict):
            raise ValueError(f"Element at '{'.'.join(keys[:keys.index(key)])}' is not a dictionary")

        if key not in current_element:
            raise KeyError(f"Key '{key}' not found at '{'.'.join(keys[:keys.index(key)])}'")

        current_element = current_element[key]

    return current_element


def set_nested_value(d: dict, value: Any, path: str) -> None:
    """
    Sets the value of a nested key in a dictionary based on the given path.

    Args:
        d: The dictionary to search.
        value: The value to set.
        path: A string representing the path to the nested key, separated by periods.

    Raises:
        ValueError: If an element in the path is not a dictionary.
        KeyError: If a key in the path does not exist.
    """
    keys = path.split(".")
    last_element = keys[-1]
    current_element = d

    for i, key in enumerate(keys):
        if key == last_element:
            current_element[key] = value
        elif key not in current_element:
            raise KeyError(f"Key '{key}' not found at '{'.'.join(keys[:keys.index(key)])}'")
        elif not isinstance(current_element[key], dict):
            raise ValueError(f"Path '{'.'.join(keys[:i + 1])}' does not lead to a dictionary.")
        else:
            current_element = current_element[key]


def format_and_raise_error(exc: Union[BaseException, subprocess.CalledProcessError]) -> None:
    """Format the error message from an exception and re-raise it as a BumpVersionError."""
    if isinstance(exc, subprocess.CalledProcessError):
        output = "\n".join([x for x in [exc.stdout, exc.stderr] if x])
        cmd = " ".join(exc.cmd)
        err_msg = f"Failed to run `{cmd}`: return code {exc.returncode}, output: {output}"
    else:  # pragma: no-coverage
        err_msg = f"Failed to run a command: {exc}"
    logger.exception(err_msg)
    raise BumpVersionError(err_msg) from exc


def run_command(command: list, env: Optional[dict] = None) -> CompletedProcess:
    """Run a shell command and return its output."""
    result = subprocess.run(  # NOQA: S603
        command, text=True, check=True, capture_output=True, env=env, encoding="utf-8"
    )
    result.check_returncode()
    return result


def is_subpath(parent: Pathlike, path: Pathlike) -> bool:
    """Return whether a path is inside the parent."""
    normalized_parent = Path(parent)
    normalized_path = Path(path)
    return str(normalized_path).startswith(str(normalized_parent)) if normalized_path.is_absolute() else True



"""Tests for the utils module."""

import sys
from itertools import combinations
from pathlib import Path

import pytest
from pytest import param

from bumpversion import utils
from bumpversion.utils import Pathlike

DRIVE_PREFIX = "c:" if sys.platform == "win32" else ""


class TestGetNestedValue:
    """Test the get_nested_value function."""

    def test_returns_value(self):
        """It should return the value of the nested key."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c"
        expected_value = 1

        # Act
        actual_value = utils.get_nested_value(d, path)

        # Assert
        assert actual_value == expected_value

    def test_invalid_path_raises_keyerror(self):
        """It should raise a KeyError if the path is invalid."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.d"

        # Act/Assert
        with pytest.raises(KeyError):
            utils.get_nested_value(d, path)

    def test_non_dict_element_raises_valueerror(self):
        """It should raise a ValueError if an element in the path is not a dict."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c.d"

        # Act/Assert
        with pytest.raises(ValueError):
            utils.get_nested_value(d, path)


class TestSetNestedValue:
    """Test the set_nested_value function."""

    def test_sets_value(self):
        """It should set the value of the nested key."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c"
        value = 2
        expected_d = {"a": {"b": {"c": 2}}}

        # Act
        utils.set_nested_value(d, value, path)

        # Assert
        assert d == expected_d

    def test_invalid_path_raises_keyerror(self):
        """It should raise a KeyError if the path is invalid."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.d.e"
        value = 2

        # Act/Assert
        with pytest.raises(KeyError):
            utils.set_nested_value(d, value, path)

    def test_non_dict_element_raises_valueerror(self):
        """It should raise a ValueError if an element in the path is not a dict."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c.d"
        value = 2

        # Act/Assert
        with pytest.raises(ValueError):
            utils.set_nested_value(d, value, path)


def pytest_generate_tests(metafunc):
    if "flag_permutation" in metafunc.fixturenames:
        flags = ["a", "i", "L", "m", "s", "u", "x"]
        permutes = flags[:]
        permutes.extend(["".join(p) for p in combinations(flags, 2)])
        permutes.extend(["".join(p) for p in combinations(flags, 3)])
        permutes.extend(["".join(p) for p in combinations(flags, 4)])
        permutes.extend(["".join(p) for p in combinations(flags, 5)])
        permutes.extend(["".join(p) for p in combinations(flags, 6)])
        permutes.extend(["".join(p) for p in combinations(flags, 7)])
        metafunc.parametrize(
            "flag_permutation",
            permutes,
        )


class TestExtractRegexFlags:
    def test_returns_flags_when_available(self, flag_permutation):
        """It should return the flags when available."""
        # Arrange
        regex = f"(?{flag_permutation})abc"
        expected = ("abc", f"(?{flag_permutation})")

        # Act
        actual_flags = utils.extract_regex_flags(regex)

        # Assert
        assert actual_flags == expected

    def test_returns_empty_string_when_no_flags(self):
        """It should return an empty string when no flags are present."""
        # Arrange
        regex = "abc"
        expected = ("abc", "")

        # Act
        actual_flags = utils.extract_regex_flags(regex)

        # Assert
        assert actual_flags == expected


@pytest.mark.parametrize(
    ["parent", "path", "expected"],
    [
        param(
            Path(f"{DRIVE_PREFIX}/parent"),
            Path(f"{DRIVE_PREFIX}/parent/subpath"),
            True,
            id="absolute_path_within_parent_returns_true",
        ),
        param(
            f"{DRIVE_PREFIX}/parent",
            Path(f"{DRIVE_PREFIX}/parent/subpath"),
            True,
            id="parent_as_string_path_as_path_1",
        ),
        param(
            Path(f"{DRIVE_PREFIX}/parent"),
            f"{DRIVE_PREFIX}/parent/subpath",
            True,
            id="parent_as_path_path_as_string_1",
        ),
        param(
            Path(f"{DRIVE_PREFIX}/parent"),
            Path(f"{DRIVE_PREFIX}/other/subpath"),
            False,
            id="absolute_path_outside_of_parent_returns_false",
        ),
        param(
            f"{DRIVE_PREFIX}/parent",
            Path(f"{DRIVE_PREFIX}/other/subpath"),
            False,
            id="parent_as_string_path_as_path_2",
        ),
        param(
            Path(f"{DRIVE_PREFIX}/parent"),
            f"{DRIVE_PREFIX}/other/subpath",
            False,
            id="parent_as_path_path_as_string_2",
        ),
        param(
            Path(f"{DRIVE_PREFIX}/parent"),
            Path(f"{DRIVE_PREFIX}/parent"),
            True,
            id="identical_absolute_paths_return_true",
        ),
        param(f"{DRIVE_PREFIX}/parent", Path(f"{DRIVE_PREFIX}/parent"), True, id="parent_as_string_path_as_path_3"),
        param(Path(f"{DRIVE_PREFIX}/parent"), f"{DRIVE_PREFIX}/parent", True, id="parent_as_path_path_as_string_3"),
        param(Path(f"{DRIVE_PREFIX}/parent"), Path("relative/path"), True, id="relative_path_returns_true"),
        param(f"{DRIVE_PREFIX}/parent", Path("relative/path"), True, id="parent_as_string_path_as_path_4"),
        param(Path(f"{DRIVE_PREFIX}/parent"), "relative/path", True, id="parent_as_path_path_as_string_4"),
    ],
)
def test_is_subpath(parent: Pathlike, path: Pathlike, expected: bool) -> None:
    """Test the is_subpath function."""
    # Act
    result = utils.is_subpath(parent, path)
    # Assert
    assert result is expected



"""Implementation of the hook interface."""

import datetime
import os
import subprocess
from typing import Dict, List, Optional

from bumpversion.config.models import Config
from bumpversion.context import get_context
from bumpversion.exceptions import HookError
from bumpversion.ui import get_indented_logger
from bumpversion.versioning.models import Version

PREFIX = "BVHOOK_"

logger = get_indented_logger(__name__)


def run_command(script: str, environment: Optional[dict] = None) -> subprocess.CompletedProcess:
    """Runs command-line programs using the shell."""
    if not isinstance(script, str):
        raise TypeError(f"`script` must be a string, not {type(script)}")
    if environment and not isinstance(environment, dict):
        raise TypeError(f"`environment` must be a dict, not {type(environment)}")
    return subprocess.run(
        script, env=environment, encoding="utf-8", shell=True, text=True, capture_output=True, check=False
    )


def base_env(config: Config) -> Dict[str, str]:
    """Provide the base environment variables."""
    return {
        f"{PREFIX}NOW": datetime.datetime.now().isoformat(),
        f"{PREFIX}UTCNOW": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        **os.environ,
        **scm_env(config),
    }


def scm_env(config: Config) -> Dict[str, str]:
    """Provide the scm environment variables."""
    scm = config.scm_info
    return {
        f"{PREFIX}COMMIT_SHA": scm.commit_sha or "",
        f"{PREFIX}DISTANCE_TO_LATEST_TAG": str(scm.distance_to_latest_tag) or "0",
        f"{PREFIX}IS_DIRTY": str(scm.dirty),
        f"{PREFIX}BRANCH_NAME": scm.branch_name or "",
        f"{PREFIX}SHORT_BRANCH_NAME": scm.short_branch_name or "",
        f"{PREFIX}CURRENT_VERSION": scm.current_version or "",
        f"{PREFIX}CURRENT_TAG": scm.current_tag or "",
    }


def version_env(version: Version, version_prefix: str) -> Dict[str, str]:
    """Provide the environment variables for each version component with a prefix."""
    return {f"{PREFIX}{version_prefix}{part.upper()}": version[part].value for part in version}


def new_version_env(config: Config, current_version: Version, new_version: Version) -> Dict[str, str]:
    """Provide the environment dictionary for new_version serialized and tag name."""
    ctx = get_context(config, current_version, new_version)
    new_version_string = config.version_config.serialize(new_version, ctx)
    ctx["new_version"] = new_version_string
    new_version_tag = config.tag_name.format(**ctx)
    return {f"{PREFIX}NEW_VERSION": new_version_string, f"{PREFIX}NEW_VERSION_TAG": new_version_tag}


def get_setup_hook_env(config: Config, current_version: Version) -> Dict[str, str]:
    """Provide the environment dictionary for `setup_hook`s."""
    return {**base_env(config), **scm_env(config), **version_env(current_version, "CURRENT_")}


def get_pre_commit_hook_env(config: Config, current_version: Version, new_version: Version) -> Dict[str, str]:
    """Provide the environment dictionary for `pre_commit_hook`s."""
    return {
        **base_env(config),
        **scm_env(config),
        **version_env(current_version, "CURRENT_"),
        **version_env(new_version, "NEW_"),
        **new_version_env(config, current_version, new_version),
    }


def get_post_commit_hook_env(config: Config, current_version: Version, new_version: Version) -> Dict[str, str]:
    """Provide the environment dictionary for `post_commit_hook`s."""
    return {
        **base_env(config),
        **scm_env(config),
        **version_env(current_version, "CURRENT_"),
        **version_env(new_version, "NEW_"),
        **new_version_env(config, current_version, new_version),
    }


def run_hooks(hooks: List[str], env: Dict[str, str], dry_run: bool = False) -> None:
    """Run a list of command-line programs using the shell."""
    logger.indent()
    for script in hooks:
        if dry_run:
            logger.info(f"Would run {script!r}")
            continue
        logger.info(f"Running {script!r}")
        logger.indent()
        result = run_command(script, env)
        if result.returncode != 0:
            logger.warning(result.stdout)
            logger.warning(result.stderr)
            raise HookError(f"{script!r} exited with {result.returncode}. ")
        else:
            logger.debug(f"Exited with {result.returncode}")
            logger.debug(result.stdout)
            logger.debug(result.stderr)
        logger.dedent()
    logger.dedent()


def run_setup_hooks(config: Config, current_version: Version, dry_run: bool = False) -> None:
    """Run the setup hooks."""
    env = get_setup_hook_env(config, current_version)
    if config.setup_hooks:
        running = "Would run" if dry_run else "Running"
        logger.info(f"{running} setup hooks:")
    else:
        logger.info("No setup hooks defined")
        return

    run_hooks(config.setup_hooks, env, dry_run)


def run_pre_commit_hooks(
    config: Config, current_version: Version, new_version: Version, dry_run: bool = False
) -> None:
    """Run the pre-commit hooks."""
    env = get_pre_commit_hook_env(config, current_version, new_version)

    if config.pre_commit_hooks:
        running = "Would run" if dry_run else "Running"
        logger.info(f"{running} pre-commit hooks:")
    else:
        logger.info("No pre-commit hooks defined")
        return

    run_hooks(config.pre_commit_hooks, env, dry_run)


def run_post_commit_hooks(
    config: Config, current_version: Version, new_version: Version, dry_run: bool = False
) -> None:
    """Run the post-commit hooks."""
    env = get_post_commit_hook_env(config, current_version, new_version)
    if config.post_commit_hooks:
        running = "Would run" if dry_run else "Running"
        logger.info(f"{running} post-commit hooks:")
    else:
        logger.info("No post-commit hooks defined")
        return

    run_hooks(config.post_commit_hooks, env, dry_run)
